import functools
import json
import traceback

from ..logging import log
from ..model.errors import HTTPError
from .coprocessor import Coprocessor, CoprocessorEvent
from .smartpipe import SmartPipe, SmartPipeEvent


def format_exception(e):
    if e is None:
        return None

    return traceback.format_exception(type(e), e, e.__traceback__)


def event_json(event):
    return {
        "id": event.id,
        "smart_pipe_id": event.smart_pipe_id,
        "input": event.input,
        "status": event.status,
        "output": event.output,
        "error": format_exception(event.error),
        "coprocessors": [
            coprocessor_event_json(coprocessor) for coprocessor in event.coprocessors
        ],
    }


def coprocessor_event_json(event):
    return {
        "id": event.id,
        "input": event.input,
        "status": event.status,
        "output": event.output,
        "error": format_exception(event.error),
    }


def coprocessor_to_json(coprocessor):
    return {"id": coprocessor.id, "type": coprocessor.type, "model": coprocessor.model}


def smart_pipe_to_json(smart_pipe):
    return {
        "id": smart_pipe.id,
        "path": smart_pipe.path,
        "method": smart_pipe.method,
        "coprocessors": [
            coprocessor_to_json(coprocessor) for coprocessor in smart_pipe.coprocessors
        ],
    }


def smart_pipes_json(smart_pipes):
    return {
        "type": "smart_pipes",
        "payload": [smart_pipe_to_json(smart_pipe) for smart_pipe in smart_pipes],
    }


class Context:
    def __init__(self, smart_pipes=[], coprocessors=[]):
        self.smart_pipes = smart_pipes
        self.coprocessors = coprocessors
        self.actual_smart_pipe_index = -1
        self.on_change = lambda x: None
        self.events = []
        self.active_event = ["none"]

    def active_smart_pipe(self, id):
        for i, smart_pipe in enumerate(self.smart_pipes):
            if smart_pipe.id == id:
                self.actual_smart_pipe_index = i
                break

    def set_event(self, event):
        self.on_change = event

    def add_event(self, event):
        self.active_event[0] = event.id
        self.events.append(event)

        self.on_change({"type": "add_request", "payload": event_json(event)})

    def update_event(self, event):
        for i, e in enumerate(self.events):
            if e.id == self.active_event[0]:
                self.events[i] = event

                self.on_change({"type": "update_request", "payload": event_json(event)})
                break

    def coprocessor_event(self, coprocessor_event):
        for i, event in enumerate(self.events):
            if event.id == self.active_event[0]:
                event.add_coprocessor_event(coprocessor_event)

                self.events[i] = event

                self.on_change({"type": "update_request", "payload": event_json(event)})
                break

    def get_actual_smart_pipe(self):
        if self.actual_smart_pipe_index == -1:
            return None

        return self.smart_pipes[self.actual_smart_pipe_index]

    def add_smart_pipe(self, smart_pipe):
        if len(self.smart_pipes) == 1 and self.smart_pipes[0].id == "temporal":
            self.smart_pipes[0] = smart_pipe
        else:
            self.actual_smart_pipe_index = len(self.smart_pipes)
            self.smart_pipes.append(smart_pipe)

        log.info(f"🧠 Smart Pipe: {smart_pipe.id}, Path: {smart_pipe.path}")
        self.on_change(smart_pipes_json(self.smart_pipes))

    def add_coprocessor(self, coprocessor):
        log.info(f"⌛️ Coprocessor {coprocessor.id} of type: {coprocessor.type}")
        self.coprocessors.append(coprocessor)

    def set_coprocessor_function(self, id, func):
        for smart_pipe in self.smart_pipes:
            for coprocessor in smart_pipe.coprocessors:
                if coprocessor.id == id:
                    coprocessor.set_func(func)
                    return coprocessor

        return None

    def assign_to_active_smart_pipe(self, coprocessor):

        self.smart_pipes[self.actual_smart_pipe_index].add_coprocessor(coprocessor)
        smart_pipe = context.get_actual_smart_pipe()
        log.info(f"⌛️ Assign Coprocessor {coprocessor.id} to Smart Pipe: {smart_pipe.id}")
        self.on_change(smart_pipes_json(self.smart_pipes))


context = Context()


def smartpipe(
    _func=None,
    path=None,
    method=None,
    id=None,
):
    def decorator_smartpipes(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            print(f"SmartPipe Path: {path}, Method: {method}, ID: {id}")
            context.active_smart_pipe(id)

            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = SmartPipeEvent(smart_pipe_id=id, input=args_json)
            context.add_event(event)
            result = None

            try:
                result = func(*args, **kwargs)
                event.set_ouput(result)
            except HTTPError as err:
                log.error(f"Smart Pipe error: {err}")
                event.set_error(err)

            context.update_event(event)
            return result

        smart_pipe = SmartPipe(wrapper, path, method, id)
        context.add_smart_pipe(smart_pipe)
        return wrapper

    if not _func:
        return decorator_smartpipes
    else:
        return decorator_smartpipes(_func)


def import_coprocessor(_func=None, coprocessor=None):
    def decorator_coprocessor(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = CoprocessorEvent(coprocessor.id, args_json)
            context.coprocessor_event(event)

            result = coprocessor.process(*args, **kwargs)

            event.set_ouput(result)
            context.coprocessor_event(event)

            return func(result)

        return wrapper

    context.add_coprocessor(coprocessor)

    if not _func:
        return decorator_coprocessor
    else:
        return decorator_coprocessor(_func)


def coprocessor(_func=None, type=None, model=None, id=None):
    def decorator_coprocessor(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            context.assign_to_active_smart_pipe(coprocessor)

            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = CoprocessorEvent(coprocessor.id, args_json)
            context.coprocessor_event(event)
            result = None

            try:
                result = coprocessor.process(*args, **kwargs)
                event.set_ouput(result)
            except HTTPError as err:
                log.error(f"Coprocessor HTTPError: {err}")
                event.set_error(err)
            except Exception as e:
                log.error(f"Coprocessor error: {e}")
                event.set_error(e)

            context.coprocessor_event(event)

            if event.error is not None:
                raise event.error

            return result

        coprocessor = Coprocessor(func, type, model, id)
        context.add_coprocessor(coprocessor)

        return wrapper

    if not _func:
        return decorator_coprocessor
    else:
        return decorator_coprocessor(_func)



import requests

from ..api.api_http import headers
from ..api.api_request import handle_request, provision_req
from ..configuration import config
from ..model.errors import SeaplaneError
from ..util import unwrap


class CoprocessorEvent:
    def __init__(self, id=None, input=None):
        self.id = id
        self.status = "in_progress"
        self.input = input
        self.output = None
        self.error = None

    def set_ouput(self, output):
        self.output = output
        self.status = "completed"

    def set_error(self, error):
        self.error = error
        self.status = "error"


SEAPLANE_API_KEY_NAME = "SEA_API_KEY"
OPENAI_API_KEY_NAME = "OPENAI_API_KEY"
REPLICATE_API_KEY_NAME = "RE_API_KEY"


def inference(version):
    def model(input_data):
        if config._api_keys is None:
            raise SeaplaneError(
                "Replicate API Key `RE_API_KEY` is not set, use `sea.config.set_api_keys`."
            )
        elif not config._api_keys.get(REPLICATE_API_KEY_NAME, None):
            raise SeaplaneError(
                "Replicate API Key `RE_API_KEY` is not set, use `sea.config.set_api_keys`."
            )

        headers = {"Authorization": f"Token {config._api_keys[REPLICATE_API_KEY_NAME]}"}

        payload = {
            "version": version,
            "input": {
                "width": 512,
                "height": 512,
                "num_outputs": 4,
                "prompt": input_data,
                "seed": 11,
            },
        }
        response = requests.post(
            "https://api.replicate.com/v1/predictions",
            json=payload,
            headers=headers,
        )

        if response.ok:
            result = response.json()
            id = result["id"]

            while not result.get("output", None):
                endpoint = f"https://api.replicate.com/v1/predictions/{id}"
                response = requests.get(
                    endpoint,
                    headers=headers,
                )

                print("Loading...")
                if response.ok:
                    result = response.json()

                    if result.get("output", None) is not None:
                        return result
        else:
            return response.text

    return model


def bloom():
    def model(input):
        url = "https://substation.dev.cplane.cloud/v1/completions"
        req = provision_req(config._token_api)

        payload = {"prompt": input}

        return unwrap(
            req(
                lambda access_token: requests.post(
                    url,
                    json=payload,
                    headers=headers(access_token),
                )
            )
        )

    return model


def gpt_3():
    def model(prompt):
        if config._api_keys is None:
            raise SeaplaneError(
                f"OPENAI API Key `{OPENAI_API_KEY_NAME}` is not set, use `sea.config.set_api_keys`."
            )
        elif not config._api_keys.get(OPENAI_API_KEY_NAME, None):
            raise SeaplaneError(
                f"OPENAI API Key `{OPENAI_API_KEY_NAME}` is not set, use `sea.config.set_api_keys`."
            )

        url = "https://api.openai.com/v1/completions"

        payload = {"model": "text-davinci-003", "prompt": prompt, "temperature": 0}

        return unwrap(
            handle_request(
                lambda: requests.post(
                    url,
                    json=payload,
                    headers=headers(config._api_keys[OPENAI_API_KEY_NAME]),
                )
            )
        )

    return model


def gpt_3_5():
    def model(input):
        if config._api_keys is None:
            raise SeaplaneError(
                f"OPENAI API Key `{OPENAI_API_KEY_NAME}` is not set, use `sea.config.set_api_keys`."
            )
        elif not config._api_keys.get(OPENAI_API_KEY_NAME, None):
            raise SeaplaneError(
                f"OPENAI API Key `{OPENAI_API_KEY_NAME}` is not set, use `sea.config.set_api_keys`."
            )

        url = "https://api.openai.com/v1/chat/completions"

        payload = {
            "model": "gpt-3.5-turbo",
            "messages": [{"role": "user", "content": input}],
            "temperature": 0.7,
        }

        return unwrap(
            handle_request(
                lambda: requests.post(
                    url,
                    json=payload,
                    headers=headers(config._api_keys[OPENAI_API_KEY_NAME]),
                )
            )
        )

    return model


class Coprocessor:
    def __init__(self, func=None, type=None, model=None, id=None, args=None, kwargs=None):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.type = type
        self.model = model
        self.id = id

    def set_func(self, func):
        self.func = func

    def process(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        print(f"Coprocessor type '{self.type}' Model ID {self.model}")

        if self.model == "bloom":
            print(f"Processing Bloom Model...")
            self.args = self.args + (bloom(),)

            return self.func(*self.args, **self.kwargs)
        elif self.model == "gpt-3.5":
            print(f"Procesing GPT-3.5 Model...")
            self.args = self.args + (gpt_3_5(),)

            return self.func(*self.args, **self.kwargs)
        elif self.model == "gpt-3":
            print(f"Procesing GPT-3 Model...")
            self.args = self.args + (gpt_3(),)

            return self.func(*self.args, **self.kwargs)
        elif self.model == "stable-diffusion":
            print(f"Accessing Replicate coprocessor...")
            self.args = self.args + (
                inference("27b93a2413e7f36cd83da926f3656280b2931564ff050bf9575f1fdf9bcd7478"),
            )

            return self.func(*self.args, **self.kwargs)
        elif self.model:
            print(f"Accessing Replicate coprocessor...")
            self.args = self.args + (inference(self.model),)

            return self.func(*self.args, **self.kwargs)
        else:
            return self.func(*self.args, **self.kwargs)

    def print(self):
        print(f"id: {self.id}, type: {self.type}, model: {self.model}")



import functools

from flask import Flask, request
from flask_cors import CORS
from flask_socketio import SocketIO, emit
from ..logging import log

from ..configuration import config
from .decorators import context, smart_pipes_json

app = Flask(__name__)
sio = SocketIO(app, cors_allowed_origins=["http://localhost:3000"], async_mode="threading")
CORS(app, origins=["http://localhost:3000"])


@sio.on("message")
def handle_message(data):
    print("received message:", data)

@sio.on("connect")
def on_connect():
    emit("message", smart_pipes_json(context.smart_pipes))


def send_something(data):
    emit("message", data, sid="sp", namespace="", broadcast=True)


def start():
    context.set_event(lambda data: send_something(data))

    smart_pipes = context.smart_pipes

    for smart_pipe in smart_pipes:

        def endpoint_func(pipe=smart_pipe):
            data = request.get_json()
            result = pipe.func(data)
            return {"result": result}

        endpoint = functools.partial(endpoint_func, pipe=smart_pipe)
        app.add_url_rule(smart_pipe.path, smart_pipe.id, endpoint, methods=[smart_pipe.method])

    def health():
        emit("message", {"data": "test"}, sid="lol", namespace="", broadcast=True)
        return "Seaplane SmartPipes Demo"

    app.add_url_rule("/", "health", health, methods=["GET"])


    if not config.is_production():        
        log.info("🚀 Smart Pipes in DEVELOPMENT MODE")
        sio.run(app, debug=False, port=1337)
    else:
        log.info("🚀 Smart Pipes in PRODUCTION")
    
    return app



import secrets
import string


def randomId():
    alphabet = string.ascii_letters + string.digits
    random_string = "".join(secrets.choice(alphabet) for i in range(10))
    return random_string


class SmartPipeEvent:
    def __init__(self, smart_pipe_id=None, input=None):
        self.id = randomId()
        self.smart_pipe_id = smart_pipe_id
        self.status = "in_progress"
        self.input = input
        self.output = None
        self.coprocessors = []
        self.error = None

    def add_coprocessor_event(self, event):
        for i, cp in enumerate(self.coprocessors):
            if cp.id == event.id:
                self.coprocessors[i] = event
                return
        self.coprocessors.append(event)

    def set_ouput(self, output):
        self.output = output
        self.status = "completed"

    def set_error(self, error):
        self.error = error
        self.status = "error"


class SmartPipe:
    def __init__(self, func=None, path=None, method=None, id=None):
        self.func = func
        self.path = path
        self.method = method
        self.id = id
        self.coprocessors = []
        self.events = []

    def add_coprocessor(self, coprocessor):
        for i, cp in enumerate(self.coprocessors):
            if cp.id == coprocessor.id:
                self.coprocessors[i] = coprocessor
                return
        self.coprocessors.append(coprocessor)
